@@ -1,4 +1,4 @@
-// Copyright 2011-2021 David Robillard <d@drobilla.net>
+// Copyright 2011-2023 David Robillard <d@drobilla.net>
 // SPDX-License-Identifier: ISC
 
 /// @file serd.h API for Serd, a lightweight RDF syntax library
@@ -12,14 +12,17 @@
 #include <stdint.h>
 #include <stdio.h>
 
-#if defined(_WIN32) && !defined(SERD_STATIC) && defined(SERD_INTERNAL)
-#  define SERD_API __declspec(dllexport)
-#elif defined(_WIN32) && !defined(SERD_STATIC)
-#  define SERD_API __declspec(dllimport)
-#elif defined(__GNUC__)
-#  define SERD_API __attribute__((visibility("default")))
-#else
-#  define SERD_API
+// SERD_API must be used to decorate things in the public API
+#ifndef SERD_API
+#  if defined(_WIN32) && !defined(SERD_STATIC) && defined(SERD_INTERNAL)
+#    define SERD_API __declspec(dllexport)
+#  elif defined(_WIN32) && !defined(SERD_STATIC)
+#    define SERD_API __declspec(dllimport)
+#  elif defined(__GNUC__)
+#    define SERD_API __attribute__((visibility("default")))
+#  else
+#    define SERD_API
+#  endif
 #endif
 
 #ifdef __GNUC__
@@ -34,19 +37,16 @@
 #  define SERD_NONNULL _Nonnull
 #  define SERD_NULLABLE _Nullable
 #  define SERD_ALLOCATED _Null_unspecified
+#  define SERD_UNSPECIFIED _Null_unspecified
 #else
 #  define SERD_NONNULL
 #  define SERD_NULLABLE
 #  define SERD_ALLOCATED
+#  define SERD_UNSPECIFIED
 #endif
 
-#define SERD_PURE_API \
-  SERD_API            \
-  SERD_PURE_FUNC
-
-#define SERD_CONST_API \
-  SERD_API             \
-  SERD_CONST_FUNC
+#define SERD_PURE_API SERD_API SERD_PURE_FUNC
+#define SERD_CONST_API SERD_API SERD_CONST_FUNC
 
 #ifndef SERD_DISABLE_DEPRECATED
 #  if defined(__clang__) && __clang_major__ >= 7
@@ -75,16 +75,16 @@ extern "C" {
 
 /// RDF syntax type
 typedef enum {
-  SERD_TURTLE   = 1, ///< Terse triples http://www.w3.org/TR/turtle
-  SERD_NTRIPLES = 2, ///< Line-based triples http://www.w3.org/TR/n-triples/
-  SERD_NQUADS   = 3, ///< Line-based quads http://www.w3.org/TR/n-quads/
-  SERD_TRIG     = 4  ///< Terse quads http://www.w3.org/TR/trig/
+  SERD_TURTLE   = 1U, ///< Terse triples http://www.w3.org/TR/turtle
+  SERD_NTRIPLES = 2U, ///< Line-based triples http://www.w3.org/TR/n-triples/
+  SERD_NQUADS   = 3U, ///< Line-based quads http://www.w3.org/TR/n-quads/
+  SERD_TRIG     = 4U, ///< Terse quads http://www.w3.org/TR/trig/
 } SerdSyntax;
 
 /// Flags indicating certain string properties relevant to serialisation
 typedef enum {
-  SERD_HAS_NEWLINE = 1u << 0u, ///< Contains line breaks ('\\n' or '\\r')
-  SERD_HAS_QUOTE   = 1u << 1u  ///< Contains quotes ('"')
+  SERD_HAS_NEWLINE = 1U << 0U, ///< Contains line breaks ('\\n' or '\\r')
+  SERD_HAS_QUOTE   = 1U << 1U, ///< Contains quotes ('"')
 } SerdNodeFlag;
 
 /// Bitwise OR of SerdNodeFlag values
@@ -103,8 +103,7 @@ typedef struct {
    library to be freed by code in the same library.  It is otherwise equivalent
    to the standard C free() function.
 */
-SERD_API
-void
+SERD_API void
 serd_free(void* SERD_NULLABLE ptr);
 
 /**
@@ -122,12 +121,13 @@ typedef enum {
   SERD_ERR_NOT_FOUND,  ///< Not found
   SERD_ERR_ID_CLASH,   ///< Encountered clashing blank node IDs
   SERD_ERR_BAD_CURIE,  ///< Invalid CURIE (e.g. prefix does not exist)
-  SERD_ERR_INTERNAL    ///< Unexpected internal error (should not happen)
+  SERD_ERR_INTERNAL,   ///< Unexpected internal error (should not happen)
+  SERD_ERR_BAD_WRITE,  ///< Error writing to file/stream
+  SERD_ERR_BAD_TEXT,   ///< Invalid text encoding
 } SerdStatus;
 
 /// Return a string describing a status code
-SERD_CONST_API
-const uint8_t* SERD_NONNULL
+SERD_CONST_API const uint8_t* SERD_NONNULL
 serd_strerror(SerdStatus status);
 
 /**
@@ -144,8 +144,7 @@ serd_strerror(SerdStatus status);
    @param n_bytes (Output) Set to the size of `str` in bytes (except NULL).
    @param flags (Output) Set to the applicable flags.
 */
-SERD_API
-size_t
+SERD_API size_t
 serd_strlen(const uint8_t* SERD_NONNULL  str,
             size_t* SERD_NULLABLE        n_bytes,
             SerdNodeFlags* SERD_NULLABLE flags);
@@ -157,10 +156,9 @@ serd_strlen(const uint8_t* SERD_NONNULL  str,
    except this function is locale-independent and always matches the lexical
    format used in the Turtle grammar (the decimal point is always ".").
 */
-SERD_API
-double
-serd_strtod(const char* SERD_NONNULL str,
-            char* SERD_NONNULL* SERD_NULLABLE endptr);
+SERD_API double
+serd_strtod(const char* SERD_NONNULL              str,
+            char* SERD_UNSPECIFIED* SERD_NULLABLE endptr);
 
 /**
    Decode a base64 string.
@@ -173,8 +171,7 @@ serd_strtod(const char* SERD_NONNULL str,
    @param size Set to the size of the returned blob in bytes.
    @return A newly allocated blob which must be freed with serd_free().
 */
-SERD_API
-void* SERD_ALLOCATED
+SERD_API void* SERD_ALLOCATED
 serd_base64_decode(const uint8_t* SERD_NONNULL str,
                    size_t                      len,
                    size_t* SERD_NONNULL        size);
@@ -192,7 +189,7 @@ serd_base64_decode(const uint8_t* SERD_NONNULL str,
 
    @return Non-zero if `stream` has encountered an error.
 */
-typedef int (*SerdStreamErrorFunc)(void* SERD_NONNULL stream);
+typedef int (*SerdStreamErrorFunc)(void* SERD_UNSPECIFIED stream);
 
 /**
    Source function for raw string input.
@@ -206,15 +203,15 @@ typedef int (*SerdStreamErrorFunc)(void* SERD_NONNULL stream);
    @param stream Stream to read from (FILE* for fread).
    @return Number of elements (bytes) read.
 */
-typedef size_t (*SerdSource)(void* SERD_NONNULL buf,
-                             size_t             size,
-                             size_t             nmemb,
-                             void* SERD_NONNULL stream);
+typedef size_t (*SerdSource)(void* SERD_NONNULL     buf,
+                             size_t                 size,
+                             size_t                 nmemb,
+                             void* SERD_UNSPECIFIED stream);
 
 /// Sink function for raw string output
 typedef size_t (*SerdSink)(const void* SERD_NONNULL buf,
                            size_t                   len,
-                           void* SERD_NONNULL       stream);
+                           void* SERD_UNSPECIFIED   stream);
 
 /**
    @}
@@ -250,9 +247,8 @@ static const SerdURI SERD_URI_NULL =
    encoding and other issues are not handled, to properly convert a file URI to
    a path, use serd_file_uri_parse().
 */
-SERD_API
 SERD_DEPRECATED_BY("serd_file_uri_parse")
-const uint8_t* SERD_NULLABLE
+SERD_API const uint8_t* SERD_NULLABLE
 serd_uri_to_path(const uint8_t* SERD_NONNULL uri);
 
 #endif
@@ -266,19 +262,16 @@ serd_uri_to_path(const uint8_t* SERD_NONNULL uri);
    @param hostname If non-NULL, set to the hostname, if present.
    @return The path component of the URI.
 */
-SERD_API
-uint8_t* SERD_NULLABLE
-serd_file_uri_parse(const uint8_t* SERD_NONNULL uri,
-                    uint8_t* SERD_NONNULL* SERD_NULLABLE hostname);
+SERD_API uint8_t* SERD_NULLABLE
+serd_file_uri_parse(const uint8_t* SERD_NONNULL              uri,
+                    uint8_t* SERD_UNSPECIFIED* SERD_NULLABLE hostname);
 
 /// Return true iff `utf8` starts with a valid URI scheme
-SERD_PURE_API
-bool
+SERD_PURE_API bool
 serd_uri_string_has_scheme(const uint8_t* SERD_NULLABLE utf8);
 
 /// Parse `utf8`, writing result to `out`
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_uri_parse(const uint8_t* SERD_NONNULL utf8, SerdURI* SERD_NONNULL out);
 
 /**
@@ -286,18 +279,16 @@ serd_uri_parse(const uint8_t* SERD_NONNULL utf8, SerdURI* SERD_NONNULL out);
 
    @see [RFC3986 5.2.2](http://tools.ietf.org/html/rfc3986#section-5.2.2)
 */
-SERD_API
-void
+SERD_API void
 serd_uri_resolve(const SerdURI* SERD_NONNULL r,
                  const SerdURI* SERD_NONNULL base,
                  SerdURI* SERD_NONNULL       t);
 
 /// Serialise `uri` with a series of calls to `sink`
-SERD_API
-size_t
+SERD_API size_t
 serd_uri_serialise(const SerdURI* SERD_NONNULL uri,
                    SerdSink SERD_NONNULL       sink,
-                   void* SERD_NONNULL          stream);
+                   void* SERD_UNSPECIFIED      stream);
 
 /**
    Serialise `uri` relative to `base` with a series of calls to `sink`
@@ -306,13 +297,12 @@ serd_uri_serialise(const SerdURI* SERD_NONNULL uri,
    `root`.  The optional `root` parameter must be a prefix of `base` and can be
    used keep up-references ("../") within a certain namespace.
 */
-SERD_API
-size_t
+SERD_API size_t
 serd_uri_serialise_relative(const SerdURI* SERD_NONNULL  uri,
                             const SerdURI* SERD_NULLABLE base,
                             const SerdURI* SERD_NULLABLE root,
                             SerdSink SERD_NONNULL        sink,
-                            void* SERD_NONNULL           stream);
+                            void* SERD_UNSPECIFIED       stream);
 
 /**
    @}
@@ -337,14 +327,14 @@ typedef enum {
 
      This type is useful as a sentinel, but is never emitted by the reader.
   */
-  SERD_NOTHING = 0,
+  SERD_NOTHING,
 
   /**
      Literal value.
 
      A literal optionally has either a language, or a datatype (not both).
   */
-  SERD_LITERAL = 1,
+  SERD_LITERAL,
 
   /**
      URI (absolute or relative).
@@ -354,7 +344,7 @@ typedef enum {
      URI (e.g. "http://example.org/foo").
      @see [RFC3986](http://tools.ietf.org/html/rfc3986)
   */
-  SERD_URI = 2,
+  SERD_URI,
 
   /**
      CURIE, a shortened URI.
@@ -362,7 +352,7 @@ typedef enum {
      Value is an unquoted CURIE string relative to the current environment,
      e.g. "rdf:type".  @see [CURIE Syntax 1.0](http://www.w3.org/TR/curie)
   */
-  SERD_CURIE = 3,
+  SERD_CURIE,
 
   /**
      A blank node.
@@ -371,7 +361,7 @@ typedef enum {
      is meaningful only within this serialisation.  @see [RDF 1.1
      Turtle](http://www.w3.org/TR/turtle/#grammar-production-BLANK_NODE_LABEL)
   */
-  SERD_BLANK = 4
+  SERD_BLANK,
 } SerdType;
 
 /// A syntactic RDF node
@@ -390,8 +380,7 @@ static const SerdNode SERD_NODE_NULL = {NULL, 0, 0, 0, SERD_NOTHING};
 
    This measures, but does not copy, `str`.  No memory is allocated.
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_from_string(SerdType type, const uint8_t* SERD_NULLABLE str);
 
 /**
@@ -400,22 +389,19 @@ serd_node_from_string(SerdType type, const uint8_t* SERD_NULLABLE str);
    This measures, but does not copy, `str`.  No memory is allocated.
    Note that the returned node may not be null terminated.
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_from_substring(SerdType                     type,
                          const uint8_t* SERD_NULLABLE str,
                          size_t                       len);
 
 /// Simple wrapper for serd_node_new_uri() to resolve a URI node
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_uri_from_node(const SerdNode* SERD_NONNULL uri_node,
                             const SerdURI* SERD_NULLABLE base,
                             SerdURI* SERD_NULLABLE       out);
 
 /// Simple wrapper for serd_node_new_uri() to resolve a URI string
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_uri_from_string(const uint8_t* SERD_NULLABLE str,
                               const SerdURI* SERD_NULLABLE base,
                               SerdURI* SERD_NULLABLE       out);
@@ -430,8 +416,7 @@ serd_node_new_uri_from_string(const uint8_t* SERD_NULLABLE str,
    If `path` is relative, `hostname` is ignored.
    If `out` is not NULL, it will be set to the parsed URI.
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_file_uri(const uint8_t* SERD_NONNULL  path,
                        const uint8_t* SERD_NULLABLE hostname,
                        SerdURI* SERD_NULLABLE       out,
@@ -447,8 +432,7 @@ serd_node_new_file_uri(const uint8_t* SERD_NONNULL  path,
    @param out Set to the parsing of the new URI (i.e. points only to
    memory owned by the new returned node).
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_uri(const SerdURI* SERD_NONNULL  uri,
                   const SerdURI* SERD_NULLABLE base,
                   SerdURI* SERD_NULLABLE       out);
@@ -465,8 +449,7 @@ serd_node_new_uri(const SerdURI* SERD_NONNULL  uri,
    @param out Set to the parsing of the new URI (i.e. points only to
    memory owned by the new returned node).
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_relative_uri(const SerdURI* SERD_NONNULL  uri,
                            const SerdURI* SERD_NULLABLE base,
                            const SerdURI* SERD_NULLABLE root,
@@ -475,8 +458,8 @@ serd_node_new_relative_uri(const SerdURI* SERD_NONNULL  uri,
 /**
    Create a new node by serialising `d` into an xsd:decimal string.
 
-   The resulting node will always contain a `.', start with a digit, and end
-   with a digit (i.e. will have a leading and/or trailing `0' if necessary).
+   The resulting node will always contain a '.', start with a digit, and end
+   with a digit (i.e. will have a leading and/or trailing '0' if necessary).
    It will never be in scientific notation.  A maximum of `frac_digits` digits
    will be written after the decimal point, but trailing zeros will
    automatically be omitted (except one if `d` is a round integer).
@@ -487,13 +470,11 @@ serd_node_new_relative_uri(const SerdURI* SERD_NONNULL  uri,
    @param d The value for the new node.
    @param frac_digits The maximum number of digits after the decimal place.
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_decimal(double d, unsigned frac_digits);
 
 /// Create a new node by serialising `i` into an xsd:integer string
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_integer(int64_t i);
 
 /**
@@ -506,8 +487,7 @@ serd_node_new_integer(int64_t i);
    @param size Size of `buf`.
    @param wrap_lines Wrap lines at 76 characters to conform to RFC 2045.
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_new_blob(const void* SERD_NONNULL buf, size_t size, bool wrap_lines);
 
 /**
@@ -515,13 +495,11 @@ serd_node_new_blob(const void* SERD_NONNULL buf, size_t size, bool wrap_lines);
 
    @return a node that the caller must free with serd_node_free().
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_node_copy(const SerdNode* SERD_NULLABLE node);
 
 /// Return true iff `a` is equal to `b`
-SERD_PURE_API
-bool
+SERD_PURE_API bool
 serd_node_equals(const SerdNode* SERD_NONNULL a,
                  const SerdNode* SERD_NONNULL b);
 
@@ -531,8 +509,7 @@ serd_node_equals(const SerdNode* SERD_NONNULL a,
    Note that if `node` is itself dynamically allocated (which is not the case
    for nodes created internally by serd), it will not be freed.
 */
-SERD_API
-void
+SERD_API void
 serd_node_free(SerdNode* SERD_NULLABLE node);
 
 /**
@@ -543,14 +520,14 @@ serd_node_free(SerdNode* SERD_NULLABLE node);
 
 /// Flags indicating inline abbreviation information for a statement
 typedef enum {
-  SERD_EMPTY_S      = 1u << 1u, ///< Empty blank node subject
-  SERD_EMPTY_O      = 1u << 2u, ///< Empty blank node object
-  SERD_ANON_S_BEGIN = 1u << 3u, ///< Start of anonymous subject
-  SERD_ANON_O_BEGIN = 1u << 4u, ///< Start of anonymous object
-  SERD_ANON_CONT    = 1u << 5u, ///< Continuation of anonymous node
-  SERD_LIST_S_BEGIN = 1u << 6u, ///< Start of list subject
-  SERD_LIST_O_BEGIN = 1u << 7u, ///< Start of list object
-  SERD_LIST_CONT    = 1u << 8u  ///< Continuation of list
+  SERD_EMPTY_S      = 1U << 1U, ///< Empty blank node subject
+  SERD_EMPTY_O      = 1U << 2U, ///< Empty blank node object
+  SERD_ANON_S_BEGIN = 1U << 3U, ///< Start of anonymous subject
+  SERD_ANON_O_BEGIN = 1U << 4U, ///< Start of anonymous object
+  SERD_ANON_CONT    = 1U << 5U, ///< Continuation of anonymous node
+  SERD_LIST_S_BEGIN = 1U << 6U, ///< Start of list subject
+  SERD_LIST_O_BEGIN = 1U << 7U, ///< Start of list object
+  SERD_LIST_CONT    = 1U << 8U, ///< Continuation of list
 } SerdStatementFlag;
 
 /// Bitwise OR of SerdStatementFlag values
@@ -572,7 +549,7 @@ typedef struct {
    @param handle Handle for user data.
    @param error Error description.
 */
-typedef SerdStatus (*SerdErrorSink)(void* SERD_NULLABLE           handle,
+typedef SerdStatus (*SerdErrorSink)(void* SERD_UNSPECIFIED        handle,
                                     const SerdError* SERD_NONNULL error);
 
 /**
@@ -580,7 +557,7 @@ typedef SerdStatus (*SerdErrorSink)(void* SERD_NULLABLE           handle,
 
    Called whenever the base URI of the serialisation changes.
 */
-typedef SerdStatus (*SerdBaseSink)(void* SERD_NULLABLE          handle,
+typedef SerdStatus (*SerdBaseSink)(void* SERD_UNSPECIFIED       handle,
                                    const SerdNode* SERD_NONNULL uri);
 
 /**
@@ -588,7 +565,7 @@ typedef SerdStatus (*SerdBaseSink)(void* SERD_NULLABLE          handle,
 
    Called whenever a prefix is defined in the serialisation.
 */
-typedef SerdStatus (*SerdPrefixSink)(void* SERD_NULLABLE          handle,
+typedef SerdStatus (*SerdPrefixSink)(void* SERD_UNSPECIFIED       handle,
                                      const SerdNode* SERD_NONNULL name,
                                      const SerdNode* SERD_NONNULL uri);
 
@@ -598,7 +575,7 @@ typedef SerdStatus (*SerdPrefixSink)(void* SERD_NULLABLE          handle,
    Called for every RDF statement in the serialisation.
 */
 typedef SerdStatus (*SerdStatementSink)(
-  void* SERD_NULLABLE           handle,
+  void* SERD_UNSPECIFIED        handle,
   SerdStatementFlags            flags,
   const SerdNode* SERD_NULLABLE graph,
   const SerdNode* SERD_NONNULL  subject,
@@ -614,7 +591,7 @@ typedef SerdStatus (*SerdStatementSink)(
    `value` will no longer be referred to by any future statements
    (i.e. the anonymous serialisation of the node is finished).
 */
-typedef SerdStatus (*SerdEndSink)(void* SERD_NULLABLE          handle,
+typedef SerdStatus (*SerdEndSink)(void* SERD_UNSPECIFIED       handle,
                                   const SerdNode* SERD_NONNULL node);
 
 /**
@@ -627,24 +604,20 @@ typedef SerdStatus (*SerdEndSink)(void* SERD_NULLABLE          handle,
 typedef struct SerdEnvImpl SerdEnv;
 
 /// Create a new environment
-SERD_API
-SerdEnv* SERD_ALLOCATED
+SERD_API SerdEnv* SERD_ALLOCATED
 serd_env_new(const SerdNode* SERD_NULLABLE base_uri);
 
 /// Free `env`
-SERD_API
-void
+SERD_API void
 serd_env_free(SerdEnv* SERD_NULLABLE env);
 
 /// Get the current base URI
-SERD_API
-const SerdNode* SERD_NONNULL
+SERD_API const SerdNode* SERD_NONNULL
 serd_env_get_base_uri(const SerdEnv* SERD_NONNULL env,
                       SerdURI* SERD_NULLABLE      out);
 
 /// Set the current base URI
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_env_set_base_uri(SerdEnv* SERD_NONNULL         env,
                       const SerdNode* SERD_NULLABLE uri);
 
@@ -655,22 +628,19 @@ serd_env_set_base_uri(SerdEnv* SERD_NONNULL         env,
    prefix "xsd" set to "http://www.w3.org/2001/XMLSchema#", "xsd:decimal" will
    expand to "http://www.w3.org/2001/XMLSchema#decimal".
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_env_set_prefix(SerdEnv* SERD_NONNULL        env,
                     const SerdNode* SERD_NONNULL name,
                     const SerdNode* SERD_NONNULL uri);
 
 /// Set a namespace prefix
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_env_set_prefix_from_strings(SerdEnv* SERD_NONNULL       env,
                                  const uint8_t* SERD_NONNULL name,
                                  const uint8_t* SERD_NONNULL uri);
 
 /// Qualify `uri` into a CURIE if possible
-SERD_API
-bool
+SERD_API bool
 serd_env_qualify(const SerdEnv* SERD_NULLABLE env,
                  const SerdNode* SERD_NONNULL uri,
                  SerdNode* SERD_NONNULL       prefix,
@@ -682,8 +652,7 @@ serd_env_qualify(const SerdEnv* SERD_NULLABLE env,
    Errors: SERD_ERR_BAD_ARG if `curie` is not valid, or SERD_ERR_BAD_CURIE if
    prefix is not defined in `env`.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_env_expand(const SerdEnv* SERD_NULLABLE env,
                 const SerdNode* SERD_NONNULL curie,
                 SerdChunk* SERD_NONNULL      uri_prefix,
@@ -694,17 +663,15 @@ serd_env_expand(const SerdEnv* SERD_NULLABLE env,
 
    Returns null if `node` can not be expanded.
 */
-SERD_API
-SerdNode
+SERD_API SerdNode
 serd_env_expand_node(const SerdEnv* SERD_NULLABLE env,
                      const SerdNode* SERD_NONNULL node);
 
 /// Call `func` for each prefix defined in `env`
-SERD_API
-void
+SERD_API void
 serd_env_foreach(const SerdEnv* SERD_NONNULL env,
                  SerdPrefixSink SERD_NONNULL func,
-                 void* SERD_NULLABLE         handle);
+                 void* SERD_UNSPECIFIED      handle);
 
 /**
    @}
@@ -716,10 +683,9 @@ serd_env_foreach(const SerdEnv* SERD_NONNULL env,
 typedef struct SerdReaderImpl SerdReader;
 
 /// Create a new RDF reader
-SERD_API
-SerdReader* SERD_ALLOCATED
-serd_reader_new(SerdSyntax          syntax,
-                void* SERD_NULLABLE handle,
+SERD_API SerdReader* SERD_ALLOCATED
+serd_reader_new(SerdSyntax             syntax,
+                void* SERD_UNSPECIFIED handle,
                 void (*SERD_NULLABLE free_handle)(void* SERD_NULLABLE),
                 SerdBaseSink SERD_NULLABLE      base_sink,
                 SerdPrefixSink SERD_NULLABLE    prefix_sink,
@@ -733,8 +699,7 @@ serd_reader_new(SerdSyntax          syntax,
    invalid characters.  Setting strict will fail when parsing such files.  An
    error is printed for invalid input in either case.
 */
-SERD_API
-void
+SERD_API void
 serd_reader_set_strict(SerdReader* SERD_NONNULL reader, bool strict);
 
 /**
@@ -743,15 +708,13 @@ serd_reader_set_strict(SerdReader* SERD_NONNULL reader, bool strict);
    The `error_sink` will be called with `handle` as its first argument.  If
    no error function is set, errors are printed to stderr in GCC style.
 */
-SERD_API
-void
+SERD_API void
 serd_reader_set_error_sink(SerdReader* SERD_NONNULL    reader,
                            SerdErrorSink SERD_NULLABLE error_sink,
-                           void* SERD_NULLABLE         error_handle);
+                           void* SERD_UNSPECIFIED      error_handle);
 
 /// Return the `handle` passed to serd_reader_new()
-SERD_PURE_API
-void* SERD_NULLABLE
+SERD_PURE_API void* SERD_UNSPECIFIED
 serd_reader_get_handle(const SerdReader* SERD_NONNULL reader);
 
 /**
@@ -763,8 +726,7 @@ serd_reader_get_handle(const SerdReader* SERD_NONNULL reader);
    corrupt data.  By setting a unique blank node prefix for each parsed file,
    this can be avoided, while preserving blank node names.
 */
-SERD_API
-void
+SERD_API void
 serd_reader_add_blank_prefix(SerdReader* SERD_NONNULL     reader,
                              const uint8_t* SERD_NULLABLE prefix);
 
@@ -775,14 +737,12 @@ serd_reader_add_blank_prefix(SerdReader* SERD_NONNULL     reader,
    node for any statements that are not in a named graph (which is currently
    all of them since Serd currently does not support any graph syntaxes).
 */
-SERD_API
-void
+SERD_API void
 serd_reader_set_default_graph(SerdReader* SERD_NONNULL      reader,
                               const SerdNode* SERD_NULLABLE graph);
 
 /// Read a file at a given `uri`
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_read_file(SerdReader* SERD_NONNULL    reader,
                       const uint8_t* SERD_NONNULL uri);
 
@@ -794,8 +754,7 @@ serd_reader_read_file(SerdReader* SERD_NONNULL    reader,
    must be ready before any callbacks will fire.  To react as soon as input
    arrives, set `bulk` to false.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_start_stream(SerdReader* SERD_NONNULL     reader,
                          FILE* SERD_NONNULL           file,
                          const uint8_t* SERD_NULLABLE name,
@@ -807,12 +766,11 @@ serd_reader_start_stream(SerdReader* SERD_NONNULL     reader,
    The `read_func` is guaranteed to only be called for `page_size` elements
    with size 1 (i.e. `page_size` bytes).
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_start_source_stream(SerdReader* SERD_NONNULL         reader,
                                 SerdSource SERD_NONNULL          read_func,
                                 SerdStreamErrorFunc SERD_NONNULL error_func,
-                                void* SERD_NONNULL               stream,
+                                void* SERD_UNSPECIFIED           stream,
                                 const uint8_t* SERD_NULLABLE     name,
                                 size_t                           page_size);
 
@@ -824,41 +782,47 @@ serd_reader_start_source_stream(SerdReader* SERD_NONNULL         reader,
    until a '.' is encountered.  This is particularly useful for reading
    directly from a pipe or socket.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_read_chunk(SerdReader* SERD_NONNULL reader);
 
 /// Finish an incremental read from a file handle
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_end_stream(SerdReader* SERD_NONNULL reader);
 
 /// Read `file`
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_read_file_handle(SerdReader* SERD_NONNULL     reader,
                              FILE* SERD_NONNULL           file,
                              const uint8_t* SERD_NULLABLE name);
 
 /// Read a user-specified byte source
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_read_source(SerdReader* SERD_NONNULL         reader,
                         SerdSource SERD_NONNULL          source,
                         SerdStreamErrorFunc SERD_NONNULL error,
-                        void* SERD_NONNULL               stream,
+                        void* SERD_UNSPECIFIED           stream,
                         const uint8_t* SERD_NULLABLE     name,
                         size_t                           page_size);
 
 /// Read `utf8`
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_reader_read_string(SerdReader* SERD_NONNULL    reader,
                         const uint8_t* SERD_NONNULL utf8);
 
+/**
+   Skip over bytes in the input until a specific byte is encountered.
+
+   Typically used for recovering from errors in a line-based syntax by skipping
+   ahead to the next newline.
+
+   @return #SERD_SUCCESS if the given byte was reached, or #SERD_FAILURE if the
+   end of input is reached.
+*/
+SERD_API SerdStatus
+serd_reader_skip_until_byte(SerdReader* SERD_NONNULL reader, uint8_t byte);
+
 /// Free `reader`
-SERD_API
-void
+SERD_API void
 serd_reader_free(SerdReader* SERD_NULLABLE reader);
 
 /**
@@ -878,31 +842,29 @@ typedef struct SerdWriterImpl SerdWriter;
    does not support abbreviation and is always ASCII.
 */
 typedef enum {
-  SERD_STYLE_ABBREVIATED = 1u << 0u, ///< Abbreviate triples when possible.
-  SERD_STYLE_ASCII       = 1u << 1u, ///< Escape all non-ASCII characters.
-  SERD_STYLE_RESOLVED    = 1u << 2u, ///< Resolve URIs against base URI.
-  SERD_STYLE_CURIED      = 1u << 3u, ///< Shorten URIs into CURIEs.
-  SERD_STYLE_BULK        = 1u << 4u, ///< Write output in pages.
+  SERD_STYLE_ABBREVIATED = 1U << 0U, ///< Abbreviate triples when possible.
+  SERD_STYLE_ASCII       = 1U << 1U, ///< Escape all non-ASCII characters.
+  SERD_STYLE_RESOLVED    = 1U << 2U, ///< Resolve URIs against base URI.
+  SERD_STYLE_CURIED      = 1U << 3U, ///< Shorten URIs into CURIEs.
+  SERD_STYLE_BULK        = 1U << 4U, ///< Write output in pages.
+  SERD_STYLE_STRICT      = 1U << 5U, ///< Abort with error on lossy output
 } SerdStyle;
 
 /// Create a new RDF writer
-SERD_API
-SerdWriter* SERD_ALLOCATED
+SERD_API SerdWriter* SERD_ALLOCATED
 serd_writer_new(SerdSyntax                   syntax,
                 SerdStyle                    style,
                 SerdEnv* SERD_NONNULL        env,
                 const SerdURI* SERD_NULLABLE base_uri,
                 SerdSink SERD_NONNULL        ssink,
-                void* SERD_NULLABLE          stream);
+                void* SERD_UNSPECIFIED       stream);
 
 /// Free `writer`
-SERD_API
-void
+SERD_API void
 serd_writer_free(SerdWriter* SERD_NULLABLE writer);
 
 /// Return the env used by `writer`
-SERD_PURE_API
-SerdEnv* SERD_NONNULL
+SERD_PURE_API SerdEnv* SERD_NONNULL
 serd_writer_get_env(SerdWriter* SERD_NONNULL writer);
 
 /**
@@ -911,11 +873,10 @@ serd_writer_get_env(SerdWriter* SERD_NONNULL writer);
    This function can be used as a SerdSink when writing to a FILE*.  The
    `stream` parameter must be a FILE* opened for writing.
 */
-SERD_API
-size_t
+SERD_API size_t
 serd_file_sink(const void* SERD_NONNULL buf,
                size_t                   len,
-               void* SERD_NONNULL       stream);
+               void* SERD_UNSPECIFIED   stream);
 
 /**
    A convenience sink function for writing to a string.
@@ -925,11 +886,10 @@ serd_file_sink(const void* SERD_NONNULL buf,
    an initialized SerdChunk.  When the write is finished, the string should be
    retrieved with serd_chunk_sink_finish().
 */
-SERD_API
-size_t
+SERD_API size_t
 serd_chunk_sink(const void* SERD_NONNULL buf,
                 size_t                   len,
-                void* SERD_NONNULL       stream);
+                void* SERD_UNSPECIFIED   stream);
 
 /**
    Finish a serialisation to a chunk with serd_chunk_sink().
@@ -937,8 +897,7 @@ serd_chunk_sink(const void* SERD_NONNULL buf,
    The returned string is the result of the serialisation, which is null
    terminated (by this function) and owned by the caller.
 */
-SERD_API
-uint8_t* SERD_NULLABLE
+SERD_API uint8_t* SERD_NULLABLE
 serd_chunk_sink_finish(SerdChunk* SERD_NONNULL stream);
 
 /**
@@ -947,11 +906,10 @@ serd_chunk_sink_finish(SerdChunk* SERD_NONNULL stream);
    The `error_sink` will be called with `handle` as its first argument.  If
    no error function is set, errors are printed to stderr.
 */
-SERD_API
-void
+SERD_API void
 serd_writer_set_error_sink(SerdWriter* SERD_NONNULL   writer,
                            SerdErrorSink SERD_NONNULL error_sink,
-                           void* SERD_NULLABLE        error_handle);
+                           void* SERD_UNSPECIFIED     error_handle);
 
 /**
    Set a prefix to be removed from matching blank node identifiers.
@@ -959,8 +917,7 @@ serd_writer_set_error_sink(SerdWriter* SERD_NONNULL   writer,
    This is the counterpart to serd_reader_add_blank_prefix() which can be used
    to "undo" added prefixes.
 */
-SERD_API
-void
+SERD_API void
 serd_writer_chop_blank_prefix(SerdWriter* SERD_NONNULL     writer,
                               const uint8_t* SERD_NULLABLE prefix);
 
@@ -969,8 +926,7 @@ serd_writer_chop_blank_prefix(SerdWriter* SERD_NONNULL     writer,
 
    Note this function can be safely casted to SerdBaseSink.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_writer_set_base_uri(SerdWriter* SERD_NONNULL      writer,
                          const SerdNode* SERD_NULLABLE uri);
 
@@ -984,8 +940,7 @@ serd_writer_set_base_uri(SerdWriter* SERD_NONNULL      writer,
    written non-relatively as <file:///foo>.  If the root is not explicitly set,
    it defaults to the base URI, so no up-references will be created at all.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_writer_set_root_uri(SerdWriter* SERD_NONNULL      writer,
                          const SerdNode* SERD_NULLABLE uri);
 
@@ -994,8 +949,7 @@ serd_writer_set_root_uri(SerdWriter* SERD_NONNULL      writer,
 
    Note this function can be safely casted to SerdPrefixSink.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_writer_set_prefix(SerdWriter* SERD_NONNULL     writer,
                        const SerdNode* SERD_NONNULL name,
                        const SerdNode* SERD_NONNULL uri);
@@ -1005,8 +959,7 @@ serd_writer_set_prefix(SerdWriter* SERD_NONNULL     writer,
 
    Note this function can be safely casted to SerdStatementSink.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_writer_write_statement(SerdWriter* SERD_NONNULL      writer,
                             SerdStatementFlags            flags,
                             const SerdNode* SERD_NULLABLE graph,
@@ -1021,8 +974,7 @@ serd_writer_write_statement(SerdWriter* SERD_NONNULL      writer,
 
    Note this function can be safely casted to SerdEndSink.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_writer_end_anon(SerdWriter* SERD_NONNULL      writer,
                      const SerdNode* SERD_NULLABLE node);
 
@@ -1032,8 +984,7 @@ serd_writer_end_anon(SerdWriter* SERD_NONNULL      writer,
    This flushes any pending output, for example terminating punctuation, so
    that the output is a complete document.
 */
-SERD_API
-SerdStatus
+SERD_API SerdStatus
 serd_writer_finish(SerdWriter* SERD_NONNULL writer);
 
 /**
