@@ -4,7 +4,7 @@
 #ifndef SERD_SRC_BYTE_SOURCE_H
 #define SERD_SRC_BYTE_SOURCE_H
 
-#include "serd/serd.h"
+#include <serd/serd.h>
 
 #include <assert.h>
 #include <stdbool.h>
@@ -58,23 +58,22 @@ SerdStatus
 serd_byte_source_page(SerdByteSource* source);
 
 static inline SERD_PURE_FUNC uint8_t
-serd_byte_source_peek(SerdByteSource* source)
+serd_byte_source_peek(SerdByteSource* const source)
 {
   assert(source->prepared);
   return source->read_buf[source->read_head];
 }
 
 static inline SerdStatus
-serd_byte_source_advance(SerdByteSource* source)
+serd_byte_source_advance(SerdByteSource* const source)
 {
   SerdStatus st = SERD_SUCCESS;
 
-  switch (serd_byte_source_peek(source)) {
-  case '\n':
+  const uint8_t c = serd_byte_source_peek(source);
+  if (c == '\n') {
     ++source->cur.line;
     source->cur.col = 0;
-    break;
-  default:
+  } else if (c) {
     ++source->cur.col;
   }
 
@@ -84,7 +83,7 @@ serd_byte_source_advance(SerdByteSource* source)
     if (source->page_size > 1) {
       if (++source->read_head == source->page_size) {
         st = serd_byte_source_page(source);
-      } else if (source->read_head == source->buf_size) {
+      } else if (source->read_head >= source->buf_size) {
         source->eof = true;
       }
     } else {
