@@ -1,8 +1,6 @@
 // Copyright 2011-2023 David Robillard <d@drobilla.net>
 // SPDX-License-Identifier: ISC
 
-#include "node.h"
-
 #include "base64.h"
 #include "string_utils.h"
 
@@ -17,6 +15,12 @@
 #include <stdlib.h>
 #include <string.h>
 
+struct SerdNodeImpl {
+  size_t        n_bytes; ///< Size in bytes (not including null)
+  SerdNodeFlags flags;   ///< Node flags (e.g. string properties)
+  SerdType      type;    ///< Node type
+};
+
 static size_t
 serd_uri_string_length(const SerdURI* const uri)
 {
@@ -288,7 +292,7 @@ serd_node_new_decimal(const double d, const unsigned frac_digits)
   char*    t   = s - 1;
   uint64_t dec = (uint64_t)int_part;
   do {
-    *t-- = (char)('0' + dec % 10);
+    *t-- = (char)('0' + (dec % 10));
   } while ((dec /= 10) > 0);
 
   *s++ = '.';
