@@ -1,10 +1,9 @@
 // Copyright 2022 David Robillard <d@drobilla.net>
 // SPDX-License-Identifier: ISC
 
-#include "serd/serd.h" // IWYU pragma: keep
+#include <serd/serd.h> // IWYU pragma: keep
 
-SERD_CONST_FUNC
-int
+SERD_CONST_FUNC int
 main(void)
 {
   return 0;
